/*
 *  SndMail by Davide Libenzi ( SMTP Internet mail delivery agent )
 *  Copyright (C) 2000, ..., 2003  Davide Libenzi
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */


#include <winsock2.h>
#include <mswsock.h>
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>
#include "Macros.h"
#include "sockio.h"
#include "BuffSock.h"
#include "Base64Enc.h"
#include "MD5.h"
#include "SndMail.h"
#include "Error.h"






static ErrorInfo USmtpErrors[] =
{
    {ERR_BAD_SMTP_RESPONSE, "Wrong SMTP server response", NULL},
    {ERR_BAD_SERVER_RESPONSE, "SMTP server error response", NULL},
    {ERR_BAD_SERVER_ADDR, "Bad SMTP server address", NULL},
    {ERR_BAD_SMTP_RESPONSE, "Bad SMTP server response", NULL},
    {ERR_FILE_CREATE, "File create error", NULL},
    {ERR_FILE_OPEN, "File open error", NULL},
    {ERR_INVALID_MSD_VERSION, "Wrong MailSendData version", NULL},
    {ERR_MSD_NO_BODYFILE, "Body file name not specified", NULL},
    {ERR_MSD_NO_SUBJECT, "Subject not specified", NULL},
    {ERR_MSD_NO_FROM, "From not specified", NULL},
    {ERR_MSD_NO_RCPT, "Rcpt not specified", NULL},
    {ERR_NO_SOCK_SUPPORT, "Socket support not found", NULL},
    {ERR_UNKNOWN_SWITCH, "Unknown switch", NULL},
    {ERR_INVALID_ERROR_CODE, "Invalid error code", NULL},
    {ERR_NO_MORE_TOKENS, "No more tokens", NULL},
    {ERR_MEMORY_ALLOCATION, "Memory allocation error", NULL},
    {ERR_NO_AUTH_INFO, "Authentication infos not supplied", NULL},
    {ERR_NETWORK, "Low level network error", NULL},
    {ERR_CONNECT, "Unable to connect", NULL},

};


static int      iErrorNo = 0;







void            ErrFreeErrorsInfo(void)
{
    for (int ii = 0; ii < ArraySize(USmtpErrors); ii++)
        if (USmtpErrors[ii].pszInfo != NULL)
            free((char *) USmtpErrors[ii].pszInfo), USmtpErrors[ii].pszInfo = NULL;

}



ErrorInfo      *ErrGetErrorInfo(int iErrorCode)
{

    for (int ii = 0; ii < ArraySize(USmtpErrors); ii++)
        if (USmtpErrors[ii].iCode == iErrorCode)
            return (&USmtpErrors[ii]);

    return (NULL);

}



char const     *ErrGetErrorString(int iErrorCode)
{

    for (int ii = 0; ii < ArraySize(USmtpErrors); ii++)
        if (USmtpErrors[ii].iCode == iErrorCode)
            return (USmtpErrors[ii].pszError);

    return ("????");

}



int             ErrSetErrorCode(int iErrorCode, char const * pszInfo)
{

    if (pszInfo != NULL)
    {
        ErrorInfo      *pEI = ErrGetErrorInfo(iErrorCode);

        if (pEI != NULL)
        {
            if (pEI->pszInfo != NULL)
                free((char *) pEI->pszInfo);

            pEI->pszInfo = strdup(pszInfo);
        }
    }

    return (iErrorNo = iErrorCode);

}



int             ErrGetErrorCode(void)
{

    return (iErrorNo);

}

